from enum import Enum


class TableStrategyTypes(Enum):
    TRUNCATE = "TRUNCATE"
    UPDATE_COLUMNS = "UPDATE_COLUMNS"

    @staticmethod
    def from_value(string):
        try:
            return TableStrategyTypes(string.upper())
        except ValueError:
            return None


# boilerplate abstract class for future use
class TableStrategy():
    pass


class TruncateTableStrategy(TableStrategy):
    strategy_type = TableStrategyTypes.TRUNCATE
    pass


class UpdateColumnsTableStrategy(TableStrategy):
    strategy_type = TableStrategyTypes.UPDATE_COLUMNS

    def __init__(self, column_strategies):
        self.column_strategies = column_strategies

    def get_column_strategies(self):
        column_strategies = {}
        for column_name, column_strategy in self.column_strategies.items():
            column_strategies[column_name] = column_strategy

        return column_strategies



from pynonymizer.output.gzip import GzipOutput
from pynonymizer.output.raw import RawOutput
import os


class UnknownOutputTypeError(Exception):
    def __init__(self, filename):
        super().__init__("Unable to detect output type for file: {}".format(filename))


def from_location(filename):
    name, ext = os.path.splitext(filename)

    if ext == ".sql":
        return RawOutput(filename)
    elif ext == ".gz":
        return GzipOutput(filename)
    else:
        raise UnknownOutputTypeError(filename)


class DatabaseStrategy:
    def __init__(self, table_strategies=None, scripts=None):

        self.table_strategies = table_strategies or {}
        self.scripts = scripts or {}

    def get_all_column_strategies(self):
        column_strategies = {}
        for table_name, table_strategy in self.table_strategies.items():
            try:
                table_columns = table_strategy.get_column_strategies()
                column_strategies.update( table_columns )
            except AttributeError as error:
                pass

        return column_strategies




from pynonymizer.strategy.exceptions import UnknownColumnStrategyError, UnknownTableStrategyError, ConfigSyntaxError
from pynonymizer.log import get_logger
from pynonymizer.strategy.table import UpdateColumnsTableStrategy, TruncateTableStrategy, TableStrategyTypes
from pynonymizer.strategy.update_column import (
    UpdateColumnStrategyTypes,
    EmptyUpdateColumnStrategy,
    UniqueEmailUpdateColumnStrategy,
    UniqueLoginUpdateColumnStrategy,
    FakeUpdateColumnStrategy,
    LiteralUpdateColumnStrategy
)
from pynonymizer.strategy.database import DatabaseStrategy
from copy import deepcopy

logger = get_logger(__name__)


class StrategyParser:
    def __init__(self, fake_seeder):
        self.fake_seeder = fake_seeder

    @staticmethod
    def __normalize_table_config(table_config):
        # If the table config doesn't have a "type" specified, it needs to be determined automatically.
        if "type" not in table_config:
            if "columns" in table_config:
                return {
                    "type": TableStrategyTypes.UPDATE_COLUMNS.value,
                    "columns": {
                        **table_config["columns"]
                    }
                }

            elif table_config == "truncate":
                return {
                    "type": TableStrategyTypes.TRUNCATE.value,
                }

            else:
                raise UnknownTableStrategyError(table_config)

        else:
            return table_config

    @staticmethod
    def __normalize_column_config(column_config):
        # If the column config doesn't have a "type" specified, it needs to be determined automatically.
        if "type" not in column_config:
            if column_config == "empty":
                return {
                    "type": UpdateColumnStrategyTypes.EMPTY.value
                }

            elif column_config == "unique_email":
                return {
                    "type": UpdateColumnStrategyTypes.UNIQUE_EMAIL.value
                }

            elif column_config == "unique_login":
                return {
                    "type": UpdateColumnStrategyTypes.UNIQUE_LOGIN.value
                }

            else:
                # Can't match it to anything special, must be a fake_update type. move the value to the fake_type field
                return {
                    "type": UpdateColumnStrategyTypes.FAKE_UPDATE.value,
                    "fake_type": column_config
                }
        else:
            return column_config

    def __parse_update_column(self, raw_column_config):
        column_config = StrategyParser.__normalize_column_config(raw_column_config)

        update_column_type = UpdateColumnStrategyTypes.from_value(column_config.pop("type"))
        try:
            if update_column_type == UpdateColumnStrategyTypes.EMPTY:
                return EmptyUpdateColumnStrategy(**column_config)

            elif update_column_type == UpdateColumnStrategyTypes.UNIQUE_LOGIN:
                return UniqueLoginUpdateColumnStrategy(**column_config)

            elif update_column_type == UpdateColumnStrategyTypes.UNIQUE_EMAIL:
                return UniqueEmailUpdateColumnStrategy(**column_config)

            elif update_column_type == UpdateColumnStrategyTypes.FAKE_UPDATE:
                return FakeUpdateColumnStrategy(self.fake_seeder, **column_config)

            elif update_column_type == UpdateColumnStrategyTypes.LITERAL:
                return LiteralUpdateColumnStrategy(**column_config)

            else:
                raise UnknownColumnStrategyError(column_config)
        except TypeError as error:
            # TypeError can be thrown when the dict args dont match the constructors for the types. We need to re-throw
            raise ConfigSyntaxError()

    def __parse_update_column_map(self,  column_map):
        parsed_columns = {}
        for name, column_config in column_map.items():
            parsed_columns[name] = self.__parse_update_column(column_config)

        return parsed_columns

    def __parse_table(self, table_config):
        table_config = StrategyParser.__normalize_table_config(table_config)

        table_strategy = TableStrategyTypes.from_value(table_config["type"])

        if table_strategy == TableStrategyTypes.TRUNCATE:
            return TruncateTableStrategy()
        elif table_strategy == TableStrategyTypes.UPDATE_COLUMNS:
            raw_column_map = table_config["columns"]
            column_strategy_map = self.__parse_update_column_map(raw_column_map)
            return UpdateColumnsTableStrategy(column_strategy_map)
        else:
            raise UnknownTableStrategyError(table_config)

    def parse_config(self, raw_config):
        """
        parse a configuration dict into a DatabaseStrategy.
        :param raw_config:
        :return:
        """
        # Deepcopy raw_config to avoid mutability issues
        config = deepcopy(raw_config)
        table_strategies = {}
        for table_name, table_config in config["tables"].items():
            table_strategies[table_name] = self.__parse_table(table_config)

        scripts = {}
        try:
            scripts = config["scripts"]
        except KeyError:
            pass

        return DatabaseStrategy(table_strategies, scripts)



import os
import uuid
from pynonymizer.database.mysql import MySqlProvider
from pynonymizer.database.exceptions import UnknownDatabaseTypeError


def get_temp_db_name(filename):
    name, _ = os.path.splitext(os.path.basename(filename))
    return f"{name}_{uuid.uuid4().hex}"


def get_provider(type, *args, **kwargs):
    if type == "mysql":
        return MySqlProvider(*args)
    else:
        raise UnknownDatabaseTypeError(type)


from enum import Enum
import yaml
from pynonymizer import input, output
from pynonymizer.log import get_default_logger
from pynonymizer.database import get_temp_db_name, get_provider
from pynonymizer.fake import FakeColumnGenerator
from pynonymizer.strategy.parser import StrategyParser
from pynonymizer.exceptions import ArgumentValidationError, DatabaseConnectionError


logger = get_default_logger()


class ProcessSteps(Enum):
    START = 0
    GET_SOURCE = 100
    CREATE_DB = 200
    RESTORE_DB = 300
    ANONYMIZE_DB = 400
    DUMP_DB = 500
    DROP_DB = 600
    END = 9999

    @staticmethod
    def names():
        return [step.name for step in ProcessSteps]

    @staticmethod
    def from_value(step_value):
        """
        resolve a enum value from key (case insensitive)
        :return: ProcessSteps property
        """
        # Try to resolve as a string value
        return ProcessSteps[step_value.upper()]


def _run_step(process_step, start_at_step, stop_at_step, skip_steps, func):
    if start_at_step.value > process_step.value:
        logger.warning(f"Skipping [{process_step.name}] (Starting at [{start_at_step.name}])")
        return False
    elif stop_at_step.value < process_step.value:
        logger.warning(f"Skipping [{process_step.name}] (Stopped at [{stop_at_step.name}])")
        return False
    elif skip_steps and process_step in skip_steps:
        logger.warning(f"Skipping [{process_step.name}] (skip-steps)")
        return False
    else:
        logger.info(f"Running [{process_step.name}]")
        func()
        return True


def pynonymize(input_path=None, strategyfile_path=None, output_path=None, db_user=None, db_password=None, db_type=None,
               db_host=None, db_name=None, fake_locale=None, start_at_step=None, stop_at_step=None, skip_steps=None):

    # Validate mandatory args
    validations = []
    if input_path is None:
        validations.append("Missing INPUT")

    if strategyfile_path is None:
        validations.append("Missing STRATEGYFILE")

    if output_path is None:
        validations.append("Missing OUTPUT")

    if db_user is None:
        validations.append("Missing DB_USER")

    if db_password is None:
        validations.append("Missing DB_PASSWORD")

    if len(validations) > 0:
        raise ArgumentValidationError(validations)

    # Default and Normalize args
    if db_type is None:
        db_type = "mysql"

    if db_host is None:
        db_host = "127.0.0.1"

    if db_name is None:
        db_name = get_temp_db_name(strategyfile_path)

    if fake_locale is None:
        fake_locale = "en_GB"

    if start_at_step is None:
        start_at_step = ProcessSteps.START
    else:
        start_at_step = ProcessSteps.from_value(start_at_step)

    if stop_at_step is None:
        stop_at_step = ProcessSteps.END
    else:
        stop_at_step = ProcessSteps.from_value(stop_at_step)

    if skip_steps and len(skip_steps) > 0:
        skip_steps = [ProcessSteps.from_value(skip) for skip in skip_steps]

    fake_seeder = FakeColumnGenerator(fake_locale)
    strategy_parser = StrategyParser(fake_seeder)

    logger.debug("loading strategyfile %s...", strategyfile_path)
    with open(strategyfile_path, "r") as strategy_yaml:
        strategy = strategy_parser.parse_config(yaml.safe_load(strategy_yaml))

    # init and validate DB connection
    logger.debug("Database: (%s)%s@%s db_name: %s", db_host, db_type, db_user, db_name)
    db_provider = get_provider(db_type, db_host, db_user, db_password, db_name)

    if not db_provider.test_connection():
        raise DatabaseConnectionError()

    # locate i/o
    input_obj = input.from_location(input_path)
    output_obj = output.from_location(output_path)
    logger.debug("input: %s output: %s", input_obj, output_obj)

    # main process
    _run_step(ProcessSteps.CREATE_DB, start_at_step, stop_at_step, skip_steps,
              lambda: db_provider.create_database())

    _run_step(ProcessSteps.RESTORE_DB, start_at_step, stop_at_step, skip_steps,
              lambda: db_provider.restore_database(input_obj))

    _run_step(ProcessSteps.ANONYMIZE_DB, start_at_step, stop_at_step, skip_steps,
              lambda: db_provider.anonymize_database(strategy))

    _run_step(ProcessSteps.DUMP_DB, start_at_step, stop_at_step, skip_steps,
              lambda: db_provider.dump_database(output_obj))

    _run_step(ProcessSteps.DROP_DB, start_at_step, stop_at_step, skip_steps,
              lambda: db_provider.drop_database())

    logger.info("Process complete!")


import subprocess
from tqdm import tqdm
from pynonymizer.database.exceptions import UnsupportedTableStrategyError
from pynonymizer.strategy.table import TableStrategyTypes
from pynonymizer.strategy.update_column import UpdateColumnStrategyTypes
from pynonymizer.database.mysql import query_factory, execution
from pynonymizer import log

class MySqlProvider:
    """
    A command-line based mysql provider. Uses `mysql` and `mysqldump`,
    Because of the efficiency of piping mass amounts of sql into the command-line client.
    Unfortunately, this implementation provides limited feedback when things go wrong.
    """
    __SEED_TABLE_NAME = "_pynonymizer_seed_fake_data"
    __CHUNK_SIZE = 8192
    __DUMPSIZE_ESTIMATE_INFLATION = 1.15
    logger = log.get_logger(__name__)

    def __init__(self, db_host, db_user, db_pass, db_name):
        self.db_host = db_host
        self.db_user = db_user
        self.db_pass = db_pass
        self.db_name = db_name
        self.__runner = execution.MySqlCmdRunner(db_host, db_user, db_pass, db_name)
        self.__dumper = execution.MySqlDumpRunner(db_host, db_user, db_pass, db_name)

    def __anonymize_table(self, table_name, table_strategy, progressbar):
        if table_strategy.strategy_type == TableStrategyTypes.TRUNCATE:
            progressbar.set_description("Truncating {}".format(table_name))
            self.__runner.db_execute(query_factory.get_truncate_table(table_name))

        elif table_strategy.strategy_type == TableStrategyTypes.UPDATE_COLUMNS:
            progressbar.set_description("Anonymizing {}".format(table_name))
            statements = query_factory.get_update_table(self.__SEED_TABLE_NAME, table_name, table_strategy.column_strategies)
            self.__runner.db_execute(statements)

        else:
            raise UnsupportedTableStrategyError(table_strategy)

        progressbar.update()

    def __seed(self, fake_update_strats, seed_rows=150):
        """
        'Seed' the database with a bunch of pre-generated random records so updates can be performed in batch updates
        """
        for i in tqdm(range(0, seed_rows), desc="Inserting seed data", unit="rows"):
            self.logger.debug(f"Inserting seed row {i}")
            self.__runner.db_execute(query_factory.get_insert_seed_row(self.__SEED_TABLE_NAME, fake_update_strats))

    def __estimate_dumpsize(self):
        """
        Makes a guess on the dump size using internal database metrics
        :return: A value in bytes, or None (unknown)
        """
        statement = query_factory.get_dumpsize_estimate(self.db_name)
        process_output = self.__runner.get_single_result(statement)

        try:
            return int(process_output) * self.__DUMPSIZE_ESTIMATE_INFLATION
        except ValueError:
            # Value unparsable, likely NULL
            return None

    def __read_until_empty_byte(self, data):
        return iter(lambda: data.read(self.__CHUNK_SIZE), b'')

    def test_connection(self):
        return self.__runner.test()

    def create_database(self):
        """
        Create the working database
        :return:
        """
        self.__runner.execute(query_factory.get_create_database(self.db_name))

    def drop_database(self):
        """
        Drop the working database
        :return:
        """
        self.__runner.execute(query_factory.get_drop_database(self.db_name))

    def anonymize_database(self, database_strategy):
        """
        Anonymize a restored database using the passed database strategy
        :param database_strategy: a strategy.DatabaseStrategy configuration
        :return:
        """
        # Filter supported columns so we're not seeding ALL types by default
        column_strats = database_strategy.get_all_column_strategies()

        fake_update_strategies = { k: v for k, v in column_strats.items() if v.strategy_type == UpdateColumnStrategyTypes.FAKE_UPDATE}

        self.logger.info("creating seed table with %d columns", len(fake_update_strategies))
        create_seed_table_sql = query_factory.get_create_seed_table(self.__SEED_TABLE_NAME, fake_update_strategies)
        self.__runner.db_execute(create_seed_table_sql)

        self.logger.info("Inserting seed data")
        self.__seed(fake_update_strategies)

        try:
            for i, before_script in enumerate(database_strategy.scripts["before"]):
                self.logger.info(f"Running before script {i} \"{before_script[:50]}\"")
                self.logger.info( self.__runner.db_execute(before_script).decode() )
        except KeyError:
            pass

        table_strategies = database_strategy.table_strategies
        self.logger.info("Anonymizing %d tables", len(table_strategies))

        with tqdm(desc="Anonymizing database", total=len(table_strategies)) as progressbar:
            for table_name, table_strategy in table_strategies.items():
                self.__anonymize_table(table_name, table_strategy, progressbar)

        try:
            for i, after_script in enumerate(database_strategy.scripts["after"]):
                self.logger.info(f"Running after script {i}: \"{after_script[:50]}\"")
                self.logger.info( self.__runner.db_execute(after_script).decode() )
        except KeyError:
            pass

        self.logger.info("dropping seed table")
        self.__runner.db_execute(query_factory.get_drop_seed_table(self.__SEED_TABLE_NAME))

    def restore_database(self, input_obj):
        """
        Feed a mysqldump dumpfile to the mysql binary on stdin.
        :param input_obj:
        :return:
        """
        dumpsize = input_obj.get_size()

        batch_processor = self.__runner.open_batch_processor()
        with input_obj.open() as dumpfile_data:
            with tqdm(desc="Restoring", total=dumpsize, unit='B', unit_scale=True, unit_divisor=1000) as bar:
                for chunk in self.__read_until_empty_byte(dumpfile_data):
                    batch_processor.write(chunk)
                    batch_processor.flush()
                    bar.update(len(chunk))

    def dump_database(self, output_obj):
        """
        Feed an output with stdout from the mysqldump binary
        :param output_obj:
        :return:
        """
        dumpsize_estimate = self.__estimate_dumpsize()

        dump_process = self.__dumper.open_dumper()
        with output_obj.open() as output_file:
            with tqdm(desc="Dumping", total=dumpsize_estimate, unit='B', unit_scale=True, unit_divisor=1000) as bar:
                for chunk in self.__read_until_empty_byte(dump_process):
                    output_file.write(chunk)
                    bar.update(len(chunk))



import struct
import gzip

class GzipInput:
    def __init__(self, filename):
        self.filename = filename

    def get_size(self):
        """
        Get the uncompressed filesize of a gzip
        The last 4 bytes of a gzip contain the uncompressed filesize in little endian format (broken on >4GB zips)
        :param filename: gzip filename to read from
        :returns uncompressed gzip filesize.
        """
        with open(self.filename, 'rb') as file:
            file.seek(-4, 2)
            size = file.read()
            return struct.unpack('<I', size)[0]

    def open(self):
        return gzip.open(self.filename, "rb")



import os

class RawInput:
    def __init__(self, filename):
        self.filename = filename

    def get_size(self):
        return os.path.getsize(self.filename)

    def open(self):
        return open(self.filename, "rb")



from pynonymizer.input.gzip import GzipInput
from pynonymizer.input.raw import RawInput
import os


class UnknownInputTypeError(Exception):
    def __init__(self, filename):
        super().__init__("Unable to detect input type for file: {}".format(filename))


def from_location(filename):
    name, ext = os.path.splitext(filename)

    if ext == ".sql":
        return RawInput(filename)
    elif ext == ".gz":
        return GzipInput(filename)
    else:
        raise UnknownInputTypeError(filename)